#include <string.h>
#include "oscardata.h"

unsigned char *aim_encode_password( const char *password, int forceLength )
{
  int i, len;
  static unsigned char encoded[16];
  
  static unsigned char encoding_table[] = {
#if 0 /* old v1 table */
    0xf3, 0xb3, 0x6c, 0x99,
    0x95, 0x3f, 0xac, 0xb6,
    0xc5, 0xfa, 0x6b, 0x63,
    0x69, 0x6c, 0xc3, 0x9f
#else /* v2.1 table, also works for ICQ */
    0xf3, 0x26, 0x81, 0xc4,
    0x39, 0x86, 0xdb, 0x92,
    0x71, 0xa3, 0xb9, 0xe6,
    0x53, 0x7a, 0x95, 0x7c
#endif
  };

  if ( !forceLength ) 
    len = strlen( password );
  else
    len = forceLength;
  
  for ( i=0; i < len; i++ )
    encoded[i] = ((unsigned char)password[i] ^ encoding_table[i]);

	return (unsigned char *)encoded;
}

void OscarData :: prepareLoginData( char *userName, char *password )
{
  unsigned char charData;
  unsigned short shortData;
  unsigned long longData;
  
  reset();
  addFLAP( 1 );
  shortData = 0;
  addData( &shortData, sizeof( unsigned short ) );
  shortData = SWAP2( 1 );
  addData( &shortData, sizeof( unsigned short ) );
  addTLV( 1, userName );
  addTLV( 2, strlen( password ), aim_encode_password( password ) );
  addTLV( 3, "Mr. Message for OS/2" );
//  addTLV( 3, "AOL Instant Messenger (SM), version 5.0.2938/WIN32" );
/*
  shortData = SWAP2( 0x0109 );
  addTLV( 0x16, 2, &shortData ); // Client ID
  shortData = SWAP2( 0x0005 );
  addTLV( 0x17, 2, &shortData ); // Major
  shortData = SWAP2( 0x0000 );
  addTLV( 0x18, 2, &shortData ); // Minor
  shortData = SWAP2( 0x0000 );
  addTLV( 0x19, 2, &shortData ); // Point
  shortData = SWAP2( 0x0b7a );
  addTLV( 0x1a, 2, &shortData ); // Build
*/
  shortData = SWAP2( 0x0500 );
  addTLV( 0x16, 2, &shortData ); // Client ID
  shortData = SWAP2( 0x0005 );
  addTLV( 0x17, 2, &shortData ); // Major
  shortData = SWAP2( 0x0000 );
  addTLV( 0x18, 2, &shortData ); // Minor
  shortData = SWAP2( 0x0000 );
  addTLV( 0x19, 2, &shortData ); // Point
  shortData = SWAP2( 0x1000 );
  addTLV( 0x1a, 2, &shortData ); // Build
  longData = SWAP4( 0x000000c7 );

  longData = SWAP4( 0x000000c7 );
  addTLV( 0x14, 4, &longData );  // distribution
  addTLV( 0x0f, "en" ); // language
  addTLV( 0x0e, "us" ); // country
//  charData = 1;
//  addTLV( 0x4a, 1, &charData ); // SSI use flag (0 = family 3, 1 = SSI only)
  tallyFLAPDataLen();
}

void OscarData :: prepareSignoffData( void )
{
  reset();
  addFLAP( 4 );
  tallyFLAPDataLen();
}

void OscarData :: prepareAuthorizationData( AuthResponseData *theData )
{
  unsigned short shortData;
  
  reset();
  addFLAP( 1 );
  shortData = 0;
  addData( &shortData, 2 );
  shortData = SWAP2( 1 );
  addData( &shortData, 2 );
  addTLV( 6, theData->authCookieLength, theData->authCookie );
  tallyFLAPDataLen();
}

void OscarData :: prepareVersionsRequest( void )
{
  SNAC theSNAC;
  unsigned short servicesRequested[] =
  { SWAP2( 1 ), SWAP2( 3 ),  // Family number followed by version.
    SWAP2( 2 ), SWAP2( 1 ),
    SWAP2( 3 ), SWAP2( 1 ),
    SWAP2( 4 ), SWAP2( 1 ),
    SWAP2( 9 ), SWAP2( 1 ),
//    SWAP2( 11 ), SWAP2( 1 ),
//    SWAP2( 13 ), SWAP2( 3 ),
//    SWAP2( 14 ), SWAP2( 1 ),
//    SWAP2( 15 ), SWAP2( 1 ),
//    SWAP2( 16 ), SWAP2( 1 ),
    SWAP2( 19 ), SWAP2( 3 ) // ,
//    SWAP2( 23 ), SWAP2( 1 )
  };
  
  // Request Generic Controls, Location Services, Buddy List Management,
  //  ICBM, Privacy Management, Usage Stats, Chat Navigation, Chat,
  //  Directory User Search, Buddy Icon Support, SSI Service,
  //  and Authorization and Registration Services
  
  reset();
  addFLAP( 2 );
  theSNAC.family = 1;
  theSNAC.subType = 0x17;
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 0x17;
  addSNAC( &theSNAC );
  
  addData( servicesRequested, 12 * sizeof( unsigned short ) );
  tallyFLAPDataLen();
}

void OscarData :: prepareReqForSelfInformation( void )
{
  SNAC theSNAC;
  
  reset();
  addFLAP( 2 );
  theSNAC.family = 1;
  theSNAC.subType = 0xe; // User information request (self)
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 0xe;
  addSNAC( &theSNAC );
  tallyFLAPDataLen();
}

void OscarData :: prepareRequestSSIlimits( void )
{
  SNAC theSNAC;
  
  reset();
  addFLAP( 2 );
  theSNAC.family = 0x13;
  theSNAC.subType = 2;
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 2;
  addSNAC( &theSNAC );
  tallyFLAPDataLen();
}

void OscarData :: prepareRateInfoRequest( void )
{
  SNAC theSNAC;
  
  reset();
  addFLAP( 2 );
  theSNAC.family = 1;
  theSNAC.subType = 6;
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 6;
  addSNAC( &theSNAC );
  tallyFLAPDataLen();
}

void OscarData :: prepareRateInformationAck( userRateInformation *rateInfo )
{
  SNAC theSNAC;
  unsigned short shortData;
  
  reset();
  addFLAP( 2 );
  theSNAC.family = 1;
  theSNAC.subType = 8;
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 8;
  addSNAC( &theSNAC );
  
  while ( rateInfo->windowSize )
  {
    shortData = SWAP2( rateInfo->rateClass );
    addData( &shortData, 2 );
    rateInfo++;
  }
  
  tallyFLAPDataLen();
}

void OscarData :: prepareCapabilitiesInfo( void )
{
  SNAC theSNAC;
  const unsigned char classInfo[] =
  {
    0x09, 0x46, 0x13, 0x43, 0x4c, 0x7f, 0x11, 0xd1, 0x82, 0x22, 0x44, 0x45,
     0x53, 0x54, 0x00, 0x00, // File transfer (send)
    0x09, 0x46, 0x13, 0x48, 0x4c, 0x7f, 0x11, 0xd1, 0x82, 0x22, 0x44, 0x45,
     0x53, 0x54, 0x00, 0x00, // File transfer (receive)
    0x09, 0x46, 0x13, 0x4b, 0x4c, 0x7f, 0x11, 0xd1, 0x82, 0x22, 0x44, 0x45,
     0x53, 0x54, 0x00, 0x00, // Buddy list transfer
    0x09, 0x46, 0x13, 0x4d, 0x4c, 0x7f, 0x11, 0xd1, 0x82, 0x22, 0x44, 0x45,
     0x53, 0x54, 0x00, 0x00, // ICQ interoperability
    0x09, 0x46, 0x13, 0x4e, 0x4c, 0x7f, 0x11, 0xd1, 0x82, 0x22, 0x44, 0x45,
     0x53, 0x54, 0x00, 0x00, // UTF-8 message support
    0x74, 0x8f, 0x24, 0x20, 0x62, 0x87, 0x11, 0xd1, 0x82, 0x22, 0x44, 0x45,
     0x53, 0x54, 0x00, 0x00, // Chat support
    0x09, 0x46, 0x13, 0x46, 0x4c, 0x7f, 0x11, 0xd1, 0x82, 0x22, 0x44, 0x45,
     0x53, 0x54, 0x00, 0x00, // Avatar (buddy icon) support
  };
  
  reset();
  addFLAP( 2 );
  theSNAC.family = 2;
  theSNAC.subType = 4;
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 4;
  addSNAC( &theSNAC );
  addTLV( 5, 112, classInfo );
  tallyFLAPDataLen();
  
  debugf( "\n" );
  debugf( "Client capabilities include:\n");
  debugf( "  File transfer (send and receive)\n" );
  debugf( "  Buddy List transfer\n" );
  debugf( "  ICQ interoperability\n" );
  debugf( "  UTF-8 support\n" );
  debugf( "  Chat support\n" );
  debugf( "  Avatar (buddy icon) support\n" );
  debugf( "\n" );
}

void OscarData :: prepareReqForSSI( void )
{
  SNAC theSNAC;
  
  reset();
  addFLAP( 2 );

  theSNAC.family = 0x13; // Server-stored information (SSI)
  theSNAC.subType = 4;   // Request all SSI
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 0x13;
  addSNAC( &theSNAC );
  tallyFLAPDataLen();
}

void OscarData :: prepareSSIActivation( void )
{
  SNAC theSNAC;
  
  reset();
  addFLAP( 2 );

  theSNAC.family = 0x13; // Server-stored information (SSI)
  theSNAC.subType = 7;   // Activate SSI data and report buddy presence info
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 0x13;
  addSNAC( &theSNAC );
  tallyFLAPDataLen();
}

void OscarData :: prepareSetStatus( void )
{
  SNAC theSNAC;
  unsigned long longData;
  
  reset();
  addFLAP( 2 );

  theSNAC.family = 1;
  theSNAC.subType = 0x1e;
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 0x1e;
  addSNAC( &theSNAC );
  longData = SWAP4( 0x01010020 );
  // Web aware, DC disabled, online, free for chat
  addTLV( 6, 4, &longData );
  
  tallyFLAPDataLen();
}

void OscarData :: prepareClientReady( const unsigned short *supportedFamilies )
{
  SNAC theSNAC;
  int i, j;
  
  typedef struct
  {
    unsigned short family, version, toolid, toolversion;
  } moduleVersionData;
  
  const moduleVersionData modVersions[] =
  {
    { SWAP2( 0x0001 ), SWAP2( 0x0003 ), SWAP2( 0x0110 ), SWAP2( 0x047b ) },// JSC
    { SWAP2( 0x0002 ), SWAP2( 0x0001 ), SWAP2( 0x0101 ), SWAP2( 0x047b ) },// JSC
    { SWAP2( 0x0003 ), SWAP2( 0x0001 ), SWAP2( 0x0110 ), SWAP2( 0x047b ) },// JSC
    { SWAP2( 0x0004 ), SWAP2( 0x0001 ), SWAP2( 0x0110 ), SWAP2( 0x047b ) },// JSC
//    { SWAP2( 0x0005 ), SWAP2( 0x0001 ), SWAP2( 0x0001 ), SWAP2( 0x0001 ) },
//    { SWAP2( 0x0006 ), SWAP2( 0x0001 ), SWAP2( 0x0110 ), SWAP2( 0x0629 ) },
//    { SWAP2( 0x0007 ), SWAP2( 0x0001 ), SWAP2( 0x0010 ), SWAP2( 0x0629 ) },
//    { SWAP2( 0x0008 ), SWAP2( 0x0001 ), SWAP2( 0x0104 ), SWAP2( 0x0001 ) },
    { SWAP2( 0x0009 ), SWAP2( 0x0001 ), SWAP2( 0x0110 ), SWAP2( 0x047b ) },// JSC
//    { SWAP2( 0x000a ), SWAP2( 0x0001 ), SWAP2( 0x0110 ), SWAP2( 0x0629 ) },
//    { SWAP2( 0x000b ), SWAP2( 0x0001 ), SWAP2( 0x0104 ), SWAP2( 0x0001 ) },
//    { SWAP2( 0x000c ), SWAP2( 0x0001 ), SWAP2( 0x0104 ), SWAP2( 0x0001 ) },
//    { SWAP2( 0x000d ), SWAP2( 0x0001 ), SWAP2( 0x0010 ), SWAP2( 0x0629 ) },
//    { SWAP2( 0x000e ), SWAP2( 0x0001 ), SWAP2( 0x0010 ), SWAP2( 0x0629 ) },
//    { SWAP2( 0x000f ), SWAP2( 0x0001 ), SWAP2( 0x0010 ), SWAP2( 0x0629 ) },
//    { SWAP2( 0x0010 ), SWAP2( 0x0001 ), SWAP2( 0x0010 ), SWAP2( 0x0629 ) },
    { SWAP2( 0x0013 ), SWAP2( 0x0003 ), SWAP2( 0x0110 ), SWAP2( 0x047b ) },// JSC
//    { SWAP2( 0x0015 ), SWAP2( 0x0001 ), SWAP2( 0x0110 ), SWAP2( 0x047c ) },
//    { SWAP2( 0x0018 ), SWAP2( 0x0001 ), SWAP2( 0x0010 ), SWAP2( 0x0629 ) },
    { 0, 0, 0, 0 } // Terminator
  };
  
  reset();
  addFLAP( 2 );
  
  theSNAC.family = 1;
  theSNAC.subType = 2;
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 2;
  addSNAC( &theSNAC );
  
  i = 0;
  while ( supportedFamilies[i] )
  {
    j = 0;
    while ( modVersions[j].family )
    {
      if ( modVersions[j].family == SWAP2( supportedFamilies[i] ) )
      {
        addData( &(modVersions[j].family), 2 );
        addData( &(modVersions[j].version), 2 );
        addData( &(modVersions[j].toolid), 2 );
        addData( &(modVersions[j].toolversion), 2 );
        break;
      }
      ++j;
    }
    ++i;
  }
  
  tallyFLAPDataLen();
}

void OscarData :: prepareSetUserInformation( char *profileString,
 char *awayMsg )
{
  SNAC theSNAC;
  const unsigned char capsBuddyIcon[16] = {
    0x09, 0x46, 0x13, 0x46, 0x4c, 0x7f, 0x11, 0xd1, 0x82, 0x22,
    0x44, 0x45, 0x53, 0x54, 0x00, 0x00
  };
  const unsigned char capsVoice[16] = {
    0x09, 0x46, 0x13, 0x41, 0x4c, 0x7f, 0x11, 0xd1, 0x82, 0x22,
    0x44, 0x45, 0x53, 0x54, 0x00, 0x00
  };
  const unsigned char capsIMImage[16] = {
    0x09, 0x46, 0x13, 0x45, 0x4c, 0x7f, 0x11, 0xd1, 0x82, 0x22,
    0x44, 0x45, 0x53, 0x54, 0x00, 0x00
  };
  const unsigned char capsChat[16] = {
    0x74, 0x8f, 0x24, 0x20, 0x62, 0x87, 0x11, 0xd1, 0x82, 0x22,
    0x44, 0x45, 0x53, 0x54, 0x00, 0x00
  };
  const unsigned char capsGetFile[16] = {
    0x09, 0x46, 0x13, 0x48, 0x4c, 0x7f, 0x11, 0xd1, 0x82, 0x22,
    0x44, 0x45, 0x53, 0x54, 0x00, 0x00
  };
  const unsigned char capsSendFile[16] = {
    0x09, 0x46, 0x13, 0x43, 0x4c, 0x7f, 0x11, 0xd1, 0x82, 0x22,
    0x44, 0x45, 0x53, 0x54, 0x00, 0x00
  };
  // All spelled out here for my future reference
  
  reset();
  addFLAP( 2 );
  
  theSNAC.family = 2;
  theSNAC.subType = 4;
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 0xfeedf00d;
  addSNAC( &theSNAC );
  
  addTLV( 1, "text/x-aolrtf; charset=\"us-ascii\"" );
  
  if ( profileString )
  {
    addTLV( 2, profileString );
  } else {
    addTLV( 2, 0, NULL );
  }
  
  addTLV( 3, "text/x-aolrtf; charset=\"us-ascii\"" );
  
  if ( awayMsg )
  {
    addTLV( 4, awayMsg );
  } else {
    addTLV( 4, 0, NULL );
  }
  
  addTLV( 5, 16, capsChat );
  // Only purport to support chat for now
  
  tallyFLAPDataLen();
}

void OscarData :: prepareRequestICBMParams( void )
{
  SNAC theSNAC;
  
  reset();
  addFLAP( 2 );

  theSNAC.family = 4;
  theSNAC.subType = 4;
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 4;
  addSNAC( &theSNAC );
  
  tallyFLAPDataLen();
}

void OscarData :: prepareSetICBMParams( int channel )
{
  SNAC theSNAC;
  const unsigned char ICBMdataChannel1[16] = {
    0x00, 0x01,             // Channel (1=IMs, 2=xfer/chat inv, 3=chat)
    0x00, 0x00, 0x00, 0x07, // Flags (support messages, missed calls,
                            //  and typing notifications)
    0x1f, 0x40,             // Max message SNAC size
    0x03, 0xe7,             // Max sender warning level
    0x03, 0xe7,             // Max receiver warning level
    0x00, 0x00,             // Minimum message interval
    0x00, 0x00,             // ??
  };
  const unsigned char ICBMdataChannel2[16] = {
    0x00, 0x02,             // Channel (1=IMs, 2=xfer/chat inv, 3=chat)
    0x00, 0x00, 0x00, 0x03, // Flags (support messages, missed calls,
                            //  and typing notifications)
    0x1f, 0x40,             // Max message SNAC size
    0x03, 0xe7,             // Max sender warning level
    0x03, 0xe7,             // Max receiver warning level
    0x00, 0x00,             // Minimum message interval
    0x00, 0x00,             // ??
  };
  const unsigned char ICBMdataChannel3[16] = {
    0x00, 0x03,             // Channel (1=IMs, 2=xfer/chat inv, 3=chat)
    0x00, 0x00, 0x00, 0x01, // Flags (support messages, missed calls,
                            //  and typing notifications)
    0x1f, 0x40,             // Max message SNAC size
    0x03, 0xe7,             // Max sender warning level
    0x03, 0xe7,             // Max receiver warning level
    0x00, 0x00,             // Minimum message interval
    0x00, 0x00,             // ??
  };
  
  reset();
  addFLAP( 2 );

  theSNAC.family = 4;
  theSNAC.subType = 2;
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 2;
  addSNAC( &theSNAC );
  
  switch ( channel )
  {
    case 1:  addData( ICBMdataChannel1, 16 ); break;
    case 2:  addData( ICBMdataChannel2, 16 ); break;
    case 3:  addData( ICBMdataChannel3, 16 ); break;
    default: reset(); return;
  }
  
  tallyFLAPDataLen();
}

void OscarData :: prepareReqLocationLimits( void )
{
  SNAC theSNAC;
  
  reset();
  addFLAP( 2 );
  
  theSNAC.family = 2;
  theSNAC.subType = 2;
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 2;
  addSNAC( &theSNAC );
  
  tallyFLAPDataLen();
}

void OscarData :: prepareReqBuddyManagementLimits( void )
{
  SNAC theSNAC;
  
  reset();
  addFLAP( 2 );
  
  theSNAC.family = 3;
  theSNAC.subType = 2;
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 2;
  addSNAC( &theSNAC );
  
  tallyFLAPDataLen();
}

void OscarData :: prepareRequestPrivacyParams( void )
{
  SNAC theSNAC;
  
  reset();
  addFLAP( 2 );

  theSNAC.family = 9;
  theSNAC.subType = 2;
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 2;
  addSNAC( &theSNAC );
  
  tallyFLAPDataLen();
}

void OscarData :: prepareRequestUserInfo( char *screenName,
 unsigned short infoType )
{
  SNAC theSNAC;
  unsigned short shortData;
  unsigned char nameLen;
  
  reset();
  addFLAP( 2 );

  theSNAC.family = 2;
  theSNAC.subType = 5;
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 5;
  addSNAC( &theSNAC );
  
  shortData = SWAP2( infoType );
  addData( &shortData, 2 );
  nameLen = strlen( screenName );
  addData( &nameLen, 1 );
  addData( screenName, nameLen );
  
  tallyFLAPDataLen();
}

void OscarData :: sendInstantMessage( const char *toUser, const char *message )
{
  SNAC theSNAC;
  unsigned long longData;
  unsigned short shortData;
  unsigned char charData;
  
  reset();
  addFLAP( 2 );

  theSNAC.family = 4;
  theSNAC.subType = 6;
  theSNAC.flags[0] = 0;
  theSNAC.flags[1] = 0;
  theSNAC.requestID = 6;
  addSNAC( &theSNAC );
  
  DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &longData, 4 );
  longData = 0;
  addData( &longData, 4 );
  longData = SWAP4( longData );
  addData( &longData, 4 );
  shortData = SWAP2( 1 );
  addData( &shortData, 2 );
  charData = strlen( toUser );
  addData( &charData, 1 );
  addData( toUser, charData );
  
  shortData = SWAP2( 2 ); 
  addData( &shortData, 2 ); // Message data TLV type
  shortData = strlen( message ) + 13;
  shortData = SWAP2( shortData );
  addData( &shortData, 2 ); // Message data TLV length
  charData = 5;
  addData( &charData, 1 );  // Fragment ID
  charData = 1;
  addData( &charData, 1 );  // Fragment version
  shortData = SWAP2( 1 );
  addData( &shortData, 2 ); // Length of capability byte array
  charData = 1;
  addData( &charData, 1 );  // Recipient must be able to read plain IM
  addData( &charData, 1 );  // Fragment ID
  addData( &charData, 1 );  // Fragment version
  shortData = strlen( message ) + 4;
  shortData = SWAP2( shortData );
  addData( &shortData, 2 ); // Length of fragment
  shortData = 0;
  addData( &shortData, 2 ); // Character set (ASCII?)
  shortData = 0xffff;
  addData( &shortData, 2 ); // Language number (all?)
  addData( message, strlen( message ) ); // THE MESSAGE!
  
  shortData = SWAP2( 6 );
  addData( &shortData, 2 ); // TLV requesting to store if recipient is offline
  shortData = 0;
  addData( &shortData, 2 ); // TLV length is 0
  
  tallyFLAPDataLen();
}

