#include <string.h>
#include <malloc.h>
#include <time.h>
#include <stdio.h>
#include <ctype.h>
#include "oscarprotocol.h"
#include "oscardata.h"

#define FREE_ALLOCD_STRING( theString ) \
  if ( theString ) { CHECKED_FREE( theString ); theString = NULL; }

AuthResponseData :: AuthResponseData()
{
  screenName = NULL;
  BOSserver = NULL;
  authCookieLength = 0;
  authCookie = NULL;
  emailAddress = NULL;
  registrationStatus = 0;
  passwordURL = NULL;
  errorCode = 0;
  errorURL = NULL;
}

AuthResponseData :: ~AuthResponseData()
{
  FREE_ALLOCD_STRING( screenName );
  FREE_ALLOCD_STRING( BOSserver );
  FREE_ALLOCD_STRING( authCookie );
  FREE_ALLOCD_STRING( emailAddress );
  FREE_ALLOCD_STRING( passwordURL );
  FREE_ALLOCD_STRING( errorURL );
}

void AuthResponseData :: printData( void )
{
  printf( "\n" );
  printf( "Authorization Response Data\n" );
  printf( "===========================\n" );
  if ( screenName )
    printf( "Screen name:                 %s\n", screenName );
  if ( BOSserver )
    printf( "BOS server:                  %s\n", BOSserver );
  if ( authCookieLength )
    printf( "Authorization cookie length: %d\n", authCookieLength );
  if ( emailAddress )
    printf( "E-mail address:              %s\n", emailAddress );
  if ( registrationStatus )
    printf( "Registration status:         %d\n", registrationStatus );
  if ( passwordURL )
    printf( "Password change URL:         %s\n", passwordURL );
  if ( errorCode )
    printf( "Error code:                  %d\n", errorCode );
  if ( errorURL )
    printf( "Error message URL:           %s\n", errorURL );
  printf( "===========================\n\n" );
}

UserInformation :: UserInformation()
{
  statusUninit = 1;
  screenName = NULL;
  warningLevel = 0;
  userClass = 0;
  userStatus = 0;
  signupDate = (time_t)0;
  signonDate = (time_t)0;
  memberSinceDate = (time_t)0;
  idleMinutes = 0;
  ipAddress[0] = 0; ipAddress[1] = 0; ipAddress[2] = 0; ipAddress[3] = 0;
  sessionLen = 0;
  buddyIconChecksum = NULL;
  buddyIconChecksumLen = 0;
  statusMessage = NULL;
  statusMessageLen = 0;
  statusMessageEncoding = NULL;
  statusMessageEncodingLen = 0;
  clientProfile = NULL;
  clientProfileLen = 0;
  clientProfileEncoding = NULL;
  clientProfileEncodingLen = 0;
  awayMessage = NULL;
  awayMessageLen = 0;
  awayMessageEncoding = NULL;
  awayMessageEncodingLen = 0;
  isOnline = 0;
}

UserInformation :: UserInformation( const UserInformation &copyMe )
{
  if ( copyMe.screenName )
  {
    CHECKED_MALLOC( strlen( copyMe.screenName ) + 1, screenName );
    strcpy( screenName, copyMe.screenName );
  }
  statusUninit = 0;
  warningLevel = copyMe.warningLevel;
  userClass = copyMe.userClass;
  userStatus = copyMe.userStatus;
  signupDate = copyMe.signupDate;
  signonDate = copyMe.signonDate;
  memberSinceDate = copyMe.memberSinceDate;
  idleMinutes = copyMe.idleMinutes;
  ipAddress[0] = copyMe.ipAddress[0];
  ipAddress[1] = copyMe.ipAddress[1];
  ipAddress[2] = copyMe.ipAddress[2];
  ipAddress[3] = copyMe.ipAddress[3];
  sessionLen = copyMe.sessionLen;
  if ( copyMe.buddyIconChecksumLen )
  {
    CHECKED_MALLOC( copyMe.buddyIconChecksumLen, buddyIconChecksum );
    buddyIconChecksumLen = copyMe.buddyIconChecksumLen;
    memcpy( buddyIconChecksum, copyMe.buddyIconChecksum, buddyIconChecksumLen );
  } else {
    buddyIconChecksumLen = 0;
    buddyIconChecksum = NULL;
  }
  if ( copyMe.statusMessageLen )
  {
    CHECKED_MALLOC( copyMe.statusMessageLen + 1, statusMessage );
    statusMessageLen = copyMe.statusMessageLen;
    strcpy( statusMessage, copyMe.statusMessage );
  } else {
    statusMessageLen = 0;
    statusMessage = NULL;
  }
  if ( copyMe.statusMessageEncodingLen )
  {
    CHECKED_MALLOC( copyMe.statusMessageEncodingLen + 1, statusMessageEncoding );
    statusMessageEncodingLen = copyMe.statusMessageEncodingLen;
    strcpy( statusMessageEncoding, copyMe.statusMessageEncoding );
  } else {
    statusMessageEncodingLen = 0;
    statusMessageEncoding = NULL;
  }
  if ( copyMe.clientProfileLen )
  {
    CHECKED_MALLOC( copyMe.clientProfileLen + 1, clientProfile );
    clientProfileLen = copyMe.clientProfileLen;
    strcpy( clientProfile, copyMe.clientProfile );
  } else {
    clientProfileLen = 0;
    clientProfile = NULL;
  }
  if ( copyMe.clientProfileEncodingLen )
  {
    CHECKED_MALLOC( copyMe.clientProfileEncodingLen + 1, clientProfileEncoding );
    clientProfileEncodingLen = copyMe.clientProfileEncodingLen;
    strcpy( clientProfileEncoding, copyMe.clientProfileEncoding );
  } else {
    clientProfileEncodingLen = 0;
    clientProfileEncoding = NULL;
  }
  if ( copyMe.awayMessageLen )
  {
    CHECKED_MALLOC( copyMe.awayMessageLen + 1, awayMessage );
    awayMessageLen = copyMe.awayMessageLen;
    strcpy( awayMessage, copyMe.awayMessage );
  } else {
    awayMessageLen = 0;
    awayMessage = NULL;
  }
  if ( copyMe.awayMessageEncodingLen )
  {
    CHECKED_MALLOC( copyMe.awayMessageEncodingLen + 1, awayMessageEncoding );
    awayMessageEncodingLen = copyMe.awayMessageEncodingLen;
    strcpy( awayMessageEncoding, copyMe.awayMessageEncoding );
  } else {
    awayMessageEncodingLen = 0;
    awayMessageEncoding = NULL;
  }
  isOnline = copyMe.isOnline;
}

void UserInformation :: incorporateData( UserInformation &copyMe )
{
  statusUninit = 0;
  
  if ( copyMe.screenName && screenName &&
        strcmp( screenName, copyMe.screenName ) )
  {
    FREE_ALLOCD_STRING( screenName );
    CHECKED_MALLOC( strlen( copyMe.screenName ) + 1, screenName );
    strcpy( screenName, copyMe.screenName );
  } else if ( !screenName && copyMe.screenName )
  {
    CHECKED_MALLOC( strlen( copyMe.screenName ) + 1, screenName );
    strcpy( screenName, copyMe.screenName );
  }
  
  if ( copyMe.warningLevel )
    warningLevel = copyMe.warningLevel;
  
  if ( copyMe.userClass )
    userClass = copyMe.userClass;
  
  if ( copyMe.userStatus )
    userStatus = copyMe.userStatus;
  
  if ( copyMe.signupDate )
    signupDate = copyMe.signupDate;
  
  if ( copyMe.signonDate )
    signonDate = copyMe.signonDate;
  
  if ( copyMe.memberSinceDate )
    memberSinceDate = copyMe.memberSinceDate;
  
  idleMinutes = copyMe.idleMinutes;
  // Always copy this so we can sense when a person re-activates after idling
  
  if ( copyMe.ipAddress[0] )
  {
    ipAddress[0] = copyMe.ipAddress[0];
    ipAddress[1] = copyMe.ipAddress[1];
    ipAddress[2] = copyMe.ipAddress[2];
    ipAddress[3] = copyMe.ipAddress[3];
  }
  
  if ( copyMe.sessionLen )
    sessionLen = copyMe.sessionLen;
  
  if ( copyMe.buddyIconChecksumLen )
  {
    FREE_ALLOCD_STRING( buddyIconChecksum );
    CHECKED_MALLOC( copyMe.buddyIconChecksumLen + 1, buddyIconChecksum );
    buddyIconChecksumLen = copyMe.buddyIconChecksumLen;
    memcpy( buddyIconChecksum, copyMe.buddyIconChecksum, buddyIconChecksumLen );
  }
  
  if ( copyMe.statusMessageLen )
  {
    FREE_ALLOCD_STRING( statusMessage );
    CHECKED_MALLOC( copyMe.statusMessageLen + 1, statusMessage );
    statusMessageLen = copyMe.statusMessageLen;
    strcpy( statusMessage, copyMe.statusMessage );
  }
  
  if ( copyMe.statusMessageEncodingLen )
  {
    FREE_ALLOCD_STRING( statusMessageEncoding );
    CHECKED_MALLOC( copyMe.statusMessageEncodingLen + 1, statusMessageEncoding );
    statusMessageEncodingLen = copyMe.statusMessageEncodingLen;
    strcpy( statusMessageEncoding, copyMe.statusMessageEncoding );
  }
  
  if ( copyMe.clientProfileLen )
  {
    FREE_ALLOCD_STRING( clientProfile );
    CHECKED_MALLOC( copyMe.clientProfileLen + 1, clientProfile );
    clientProfileLen = copyMe.clientProfileLen;
    strcpy( clientProfile, copyMe.clientProfile );
  }
  
  if ( copyMe.clientProfileEncodingLen )
  {
    FREE_ALLOCD_STRING( clientProfileEncoding );
    CHECKED_MALLOC( copyMe.clientProfileEncodingLen + 1, clientProfileEncoding );
    clientProfileEncodingLen = copyMe.clientProfileEncodingLen;
    strcpy( clientProfileEncoding, copyMe.clientProfileEncoding );
    if ( awayMessageLen )
    {
      FREE_ALLOCD_STRING( awayMessage );
      awayMessageLen = 0;
    }
  }
  
  if ( copyMe.awayMessageLen )
  {
    FREE_ALLOCD_STRING( awayMessage );
    CHECKED_MALLOC( copyMe.awayMessageLen + 1, awayMessage );
    awayMessageLen = copyMe.awayMessageLen;
    strcpy( awayMessage, copyMe.awayMessage );
  }
  
  if ( copyMe.awayMessageEncodingLen )
  {
    FREE_ALLOCD_STRING( awayMessageEncoding );
    CHECKED_MALLOC( copyMe.awayMessageEncodingLen + 1, awayMessageEncoding );
    awayMessageEncodingLen = copyMe.awayMessageEncodingLen;
    strcpy( awayMessageEncoding, copyMe.awayMessageEncoding );
  }
  
  isOnline = copyMe.isOnline;
}

UserInformation :: ~UserInformation()
{
  FREE_ALLOCD_STRING( screenName );
  FREE_ALLOCD_STRING( buddyIconChecksum );
  FREE_ALLOCD_STRING( statusMessage );
  FREE_ALLOCD_STRING( statusMessageEncoding );
  FREE_ALLOCD_STRING( clientProfile );
  FREE_ALLOCD_STRING( clientProfileEncoding );
  FREE_ALLOCD_STRING( awayMessage );
  FREE_ALLOCD_STRING( awayMessageEncoding );
}

#define PRINT_USERCLASS_LINE( bit, classString )   \
    if ( tmpLong & bit )                           \
    {                                              \
      printf( classString );                       \
      tmpLong &= ~bit;                             \
      if ( tmpLong ) printf( "                " ); \
    }


void UserInformation :: printData( void )
{
  unsigned long tmpLong;
  
  printf( "\n" );
  printf( "User Information Data\n" );
  printf( "=====================\n" );
  if ( screenName )
    printf( "Screen name:    %s\n", screenName );
  if ( warningLevel )
    printf( "Warning level:  %d\n", warningLevel );
  if ( userClass )
  {
    tmpLong = userClass & 4095;
    
    printf( "User class:     " );
    PRINT_USERCLASS_LINE( 1, "Trial (user less than 60 days)\n" );
    PRINT_USERCLASS_LINE( 2, "Administrator\n" );
    PRINT_USERCLASS_LINE( 4, "AOL Lemming\n" );
    PRINT_USERCLASS_LINE( 8, "Commercial Account\n" );
    PRINT_USERCLASS_LINE( 16, "Free Account\n" );
    PRINT_USERCLASS_LINE( 32, "Away\n" );
    PRINT_USERCLASS_LINE( 64, "ICQ User\n" );
    PRINT_USERCLASS_LINE( 128, "Wireless User\n" );
    PRINT_USERCLASS_LINE( 256, "UNKNOWN USER CLASS (0x100)\n" );
    PRINT_USERCLASS_LINE( 512, "UNKNOWN USER CLASS (0x200)\n" );
    PRINT_USERCLASS_LINE( 1024, "ActiveBuddy Account\n" );
    PRINT_USERCLASS_LINE( 2048, "UNKNOWN USER CLASS (0x800)\n" );
  }
  if ( userStatus )
  {
    tmpLong = userStatus & 0x312b0136; // Only the bits I recognize
    
    printf( "User status:    " );
    PRINT_USERCLASS_LINE( 0x00000001, "Away\n" );
    PRINT_USERCLASS_LINE( 0x00000002, "Do Not Disturb\n" );
    PRINT_USERCLASS_LINE( 0x00000004, "Not Available\n" );
    PRINT_USERCLASS_LINE( 0x00000010, "Occupied\n" );
    PRINT_USERCLASS_LINE( 0x00000020, "Free For Chat\n" );
    PRINT_USERCLASS_LINE( 0x00000100, "Invisible\n" );
    PRINT_USERCLASS_LINE( 0x00010000, "Web Aware\n" );
    PRINT_USERCLASS_LINE( 0x00020000, "Visible IP\n" );
    PRINT_USERCLASS_LINE( 0x00080000, "Birthday\n" );
    PRINT_USERCLASS_LINE( 0x00200000, "Web-front\n" );
    PRINT_USERCLASS_LINE( 0x01000000, "Direct Connection disabled\n" );
    PRINT_USERCLASS_LINE( 0x10000000, "Direct Connection on authorization\n" );
    PRINT_USERCLASS_LINE( 0x20000000, "Direct Connection for contacts\n" );
  }
  
  if ( signupDate )
    printf( "Signup date:    %s", ctime( &signupDate ) );
  if ( signonDate )
    printf( "Signon date:    %s", ctime( &signonDate ) );
  if ( memberSinceDate )
    printf( "Member since:   %s", ctime( &memberSinceDate ) );
  if ( idleMinutes )
    printf( "Idle time:      %ld minutes\n", idleMinutes );
  if ( ipAddress[0] )
    printf( "IP address:     %d.%d.%d.%d\n", ipAddress[0], ipAddress[1],
     ipAddress[2], ipAddress[3] );
  if ( sessionLen )
    printf( "Session length: %ld\n", sessionLen );
  if ( statusMessageLen )
    printf( "Status message: %s\n", statusMessage );
  if ( statusMessageEncodingLen )
    printf( "Status message encoding: %s\n", statusMessageEncoding );
  if ( buddyIconChecksumLen )
    printf( "Received buddy icon checksum (%d bytes).\n", buddyIconChecksumLen );
  if ( clientProfileLen )
    printf( "Client profile:          %s\n", clientProfile );
  if ( clientProfileEncodingLen )
    printf( "Client profile encoding: %s\n", clientProfileEncoding );
  if ( awayMessageLen )
    printf( "Away message:            %s\n", awayMessage );
  if ( awayMessageEncoding )
    printf( "Away message encoding:   %s\n", awayMessageEncoding );
  printf( "=====================\n\n" );
}

#define STRCAT( stringToAdd ) \
{ \
  addStrLen = strlen( stringToAdd ); \
  if ( addStrLen ) \
  { \
    if ( (strLen + addStrLen + 1) > allocLen ) \
    { \
      while ( (strLen + addStrLen + 1) > allocLen ) allocLen *= 2; \
      CHECKED_REALLOC( buffer, allocLen, buffer ); \
    } \
    strcpy( buffer + strLen, (char *)stringToAdd ); \
    strLen += addStrLen; \
  } \
}

char *stripTags( char *origString )
{
  char *retString;
  int i, editStart, len;
  
  len = strlen( origString );
  CHECKED_MALLOC( len + 1, retString );
  strcpy( retString, origString );
  
  for ( i=0; i < len; ++i )
  {
    if ( i < (len - 4) && strnicmp( retString + i, "&gt;", 4 ) == 0 )
    {
      retString[i] = '>';
      strcpy( retString + i + 1, retString + i + 4 );
      len -= 3;
      continue;
    }
    // Replace &gt; with a great than sign
    
    if ( i < (len - 4) && strnicmp( retString + i, "&lt;", 4 ) == 0 )
    {
      retString[i] = '<';
      strcpy( retString + i + 1, retString + i + 4 );
      len -= 3;
      continue;
    }
    // Replace &lt; with a less than sign
    
    if ( i < (len - 6) && strnicmp( retString + i, "&quot;", 6 ) == 0 )
    {
      retString[i] = '"';
      strcpy( retString + i + 1, retString + i + 6 );
      len -= 5;
      continue;
    }
    // Replace &quot; with a quote
    
    if ( i < (len - 5) && strnicmp( retString + i, "&amp;", 5 ) == 0 )
    {
      retString[i] = '"';
      strcpy( retString + i + 1, retString + i + 5 );
      len -= 4;
      continue;
    }
    // Replace &quot; with a quote
    
    if ( retString[i] == '<' )
    {
      editStart = i;
      while ( i < len && retString[i] != '>' ) ++i;
      if ( retString[i] == '>' )
      {
        retString[i] = 0;
        if ( stricmp( retString + editStart + 1, "br" ) == 0 )
        {
          retString[editStart] = '\n';
          editStart++;
          len++;
        }
        // Replace <BR> tag with a \n
        
        strcpy( retString + editStart, retString + i + 1 );
        len -= (i - editStart) + 1;
        i = editStart - 1;
        continue;
      } else {
        retString[editStart] = 0;
        return retString;
      }
    }
  }
  return retString;
}

char *UserInformation :: getUserBlurbString( void )
{
  char *buffer;
  char buffer2[16];
  int allocLen, strLen, addStrLen;
  
  allocLen = 128;
  CHECKED_MALLOC( 128, buffer );
  strLen = 0;
  
  if ( !screenName || !isOnline )
  {
    STRCAT( "Not signed on." );
    return buffer;
  }
  
  STRCAT( "Screen Name: " );
  STRCAT( screenName );
  
  if ( signonDate )
  {
    STRCAT( "\nSigned on: " );
    STRCAT( ctime( &signonDate ) );
    strLen--;
    buffer[ strLen ] = 0;
    // Hack off the newline
  }
  if ( idleMinutes )
  {
    STRCAT( "\nIdle (hours:mins): " );
    sprintf( buffer2, "%d:%02d", idleMinutes / 60, idleMinutes % 60 );
    STRCAT( buffer2 );
  }
  if ( clientProfileLen )
  {
    char *tmpString;
    STRCAT( "\nUser profile:\n" );
    tmpString = stripTags( clientProfile );
    STRCAT( tmpString );
    CHECKED_FREE( tmpString );
  }
  if ( awayMessageLen )
  {
    char *tmpString;
    STRCAT( "\nAway message:\n" );
    tmpString = stripTags( awayMessage );
    STRCAT( tmpString );
    CHECKED_FREE( tmpString );
  }

  return buffer;
}

SSIData :: SSIData()
{
  numRootBuddies = 0;
  rootBuddies = NULL;
  dateStamp = (time_t) 0;
}

SSIData :: ~SSIData()
{
  int i;
  
  for ( i=0; i<numRootBuddies; ++i )
  {
    if ( rootBuddies[i].entryName )
    {
      CHECKED_FREE( rootBuddies[i].entryName );
    }
    if ( rootBuddies[i].memberIDs )
    {
      CHECKED_FREE( rootBuddies[i].memberIDs );
    }
  }
  
  CHECKED_FREE( rootBuddies );
}

void SSIData :: printData( void )
{
  int i, j, k, needSkip;
  
  printf( "\n" );
  printf( "Server Stored Information (Buddy List)\n" );
  printf( "======================================\n" );

  needSkip = 0;
  for ( i=0; i<numRootBuddies; ++i )
  {
    if ( rootBuddies[i].numMembers )
    {
      if ( needSkip )
      {
        printf( "\n" );
      } else needSkip = 1;
      
      rootBuddies[i].beenHere = 1;
      printf( "Group Name:    %s\n", rootBuddies[i].entryName );
      printf( "Group ID:      %d\n", rootBuddies[i].gid );
      printf( "Group Members: " );
      if ( !rootBuddies[i].numMembers )
      {
        printf( "<NONE>\n" );
      } else for ( j=0; j<rootBuddies[i].numMembers; ++j )
      {
        for ( k=0; k<numRootBuddies; ++k )
        {
          if ( rootBuddies[k].id == rootBuddies[i].memberIDs[j] &&
                rootBuddies[k].gid == rootBuddies[i].gid )
          {
            // Found the right entry
            if ( j )
            {
              printf( "               " );
            }
            printf( "%s\n", rootBuddies[k].entryName );
            rootBuddies[k].beenHere = 1;
          }
        }
      }
    }
  }
  
  for ( i=0; i<numRootBuddies; ++i )
  {
    if ( !rootBuddies[i].beenHere )
    {
      if ( needSkip )
      {
        needSkip = 0;
        printf( "\n" );
      }
      printf( "Orphan Member: %s\n", rootBuddies[i].entryName );
    } else {
      rootBuddies[i].beenHere = 0;
      // Reset the status to be neat about it
    }
  }
  
  if ( dateStamp )
  {
    if ( numRootBuddies )
      printf( "\n" );
    
    printf( "Last update:   %s", ctime( &dateStamp ) );
  }
  
  printf( "======================================\n\n" );
}

