#ifndef OSCARSESSION_H_INCLUDED
#define OSCARSESSION_H_INCLUDED

#include "oscardata.h"
#include "buddy.h"
#include "errors.h"

typedef enum
{
  noSession, oscarConnectionAck, oscarLoginAuthorization, BOSconnectionAck,
  BOShostReady, BOSservicesVersion, BOScurrentUserInfo, BOSrateInformation,
  BOSserverStoredLimits, BOSserverStoredInfo, BOSlocationLimits,
  BOSbuddyManagementLimits, BOSprivacyLimits, BOSICBMparams, BOSserviceLoop,
  shutDown, nextSocket, fatalError
} oscarSessionState;

class OscarSession
{
protected:
  //
  // Network-related data
  //
  
  oscarSessionState status, waitForState;
  // status is where we are now
  // waitForStatus is the state we want to have the stateWakeup sem posted for
  int currentSequence;
  OscarData sendData, receiveData;
  int oscarSocket, BOSsocket;
  AuthResponseData authData;
  UserInformation userInfo;
  SSIData ssiData;
  
  //
  // GUI-related data
  //
  
  HWND sessionWindow;
  Buddy **buddyList;
  int numBuddies;
  MINIRECORDCORE *myRecordInManager;
  
  //
  // Thread management
  //
  
  ULONG dataReceiverThreadID;
  HEV stateWakeup, continuousReceive, sessionShutdown;
  HMTX waitStateMux;
  
  //
  // Internal data
  //
  
  errorStatus errorCode;

public:
  unsigned short *supportedFamilies;
  userRateInformation *rateInformation;

  OscarSession( char *userName, char *password, HWND mgrWin,
   MINIRECORDCORE *theRecord, OscarSession **storeHere );
  // Username and password must be allocated outside of this function and will
  //  be freed inside of this function.  Do not use these pointers in your
  //  code after returning from here.
  
  ~OscarSession();
  
  errorStatus getErrorStatus( void ) const;
  oscarSessionState getState( void ) const;
  
  void setState( oscarSessionState newState );
  
  Buddy **getBuddyList( int *numBuds );
  HWND getBuddyListWindow( void ) const;
  int *getSeqNumPointer( void );
  const char *getUserName( void ) const;
  
  void queueForSend( OscarData *queueData );
  
  void serviceSessionUntilClosed( void );
  void closeSession( void );
  void windowClosed( void );
  
  void reportAvailableServices( void );
  
  friend void dataReceiverThread( OscarSession *sessionData );
  friend void rateRegulationThread( OscarSession *sessionData );
  // Allows the threads to touch my naughty-bits
};

void dataReceiverThread( OscarSession *sessionData );
// Handles all data received from Oscar/BOS server

void rateRegulationThread( OscarSession *sessionData );
// Controls the rate at which certain messages are sent to avoid flooding
//  the AOL servers

inline errorStatus OscarSession :: getErrorStatus( void ) const
{
  return errorCode;
}

inline oscarSessionState OscarSession :: getState( void ) const
{
  return status;
}

inline void OscarSession :: setState( oscarSessionState newState )
{
  status = newState;
}

inline Buddy **OscarSession :: getBuddyList( int *numBuds )
{
  *numBuds = numBuddies;
  return buddyList;
}

inline HWND OscarSession :: getBuddyListWindow( void ) const
{
  return sessionWindow;
}

inline int *OscarSession :: getSeqNumPointer( void )
{
  return &currentSequence;
}

inline const char *OscarSession :: getUserName( void ) const
{
  return userInfo.screenName;
}

#endif
